@@ -27,6 +27,74 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: main.cxx,v $
+ * Revision 1.84  2002/11/13 10:23:12  rogerh
+ * Enable Speex support by default.
+ *
+ * Revision 1.83  2002/11/10 08:12:42  robertj
+ * Moved constants for "well known" ports to better place (OPAL change).
+ *
+ * Revision 1.82  2002/08/21 08:48:09  rogerh
+ * Fix some problems caused by the user specif OGM changes
+ *
+ * Revision 1.81  2002/08/21 06:55:09  rogerh
+ * Support all 5 Speex bitrates.
+ *
+ * Revision 1.80  2002/08/16 02:04:30  craigs
+ * Ensure compilation without Speex installed
+ *
+ * Revision 1.79  2002/08/15 18:52:49  rogerh
+ * Add support for one of the Speex codecs
+ *
+ * Revision 1.78  2002/08/05 10:06:06  robertj
+ * Changed to use the version of G.7231. file capability/codec as used in
+ *   opalvxml module. Prevents duplicate symbol link errors.
+ *
+ * Revision 1.77  2002/06/28 00:30:41  robertj
+ * Fixed Clone() for G.723 file capability, must clone contents!
+ *
+ * Revision 1.76  2002/05/08 03:15:35  robertj
+ * Removed unecessary initialisation of frameBuffer PBYTEArray
+ *
+ * Revision 1.75  2002/04/01 14:45:43  craigs
+ * Fixed flag to end call after OGM
+ *
+ * Revision 1.74  2002/04/01 13:16:09  craigs
+ * Added ability to kill calls at end of playing OGM
+ * Added ability to play different OGMs depending on target DN
+ *
+ * Revision 1.73  2002/02/21 07:21:52  rogerh
+ * Change the outgoing audio's FrameDelay() code to use the PAdaptiveDelay
+ * class which should give clearer audio for the OGMs.
+ *
+ * Revision 1.72  2002/02/04 13:23:24  rogerh
+ * Change the silence frame back to a 4 byte SID. It caused Quicnet cards
+ * to play a loud noise instead of silence.
+ *
+ * Revision 1.71  2002/01/31 16:12:26  rogerh
+ * Clear the end of the frame buffer if we cannot read a full frame.
+ *
+ * Revision 1.70  2002/01/26 09:40:03  rogerh
+ * Make OpenAM send out G.723.1 OGM audio files correctly.
+ *
+ * Revision 1.69  2002/01/25 11:50:10  rogerh
+ * dd --port option to play and record commands so you can use a
+ * telephone on the POTS port or use speakers and microphone.
+ *
+ * Revision 1.68  2002/01/25 08:58:46  rogerh
+ * Send out silence using a standard 24 byte frame instead of a 4 byte SID.
+ * This is a workaround to support an end point which does not understand SID.
+ *
+ * Revision 1.67  2002/01/22 04:15:04  craigs
+ * Updated for move of PWavFile to PTCLib
+ * Replaced AudioDelay with PAdaptiveDelay
+ *
+ * Revision 1.66  2002/01/11 16:07:15  rogerh
+ * Make PCMExt non static as it is used externally in cmds.cxx
+ *
+ * Revision 1.65  2001/11/18 23:07:36  craigs
+ * Fixed problem whereby recording could start when queue of OGMs was not empty
+ * Thanks to Frank Derks
+ *
  * Revision 1.64  2001/10/24 10:45:33  rogerh
  * replace 4 byte SID frames with 24 byte frames of silence. Windows Media
  * Player's codec cannot handle SID frames.
@@ -247,7 +315,10 @@
 
 #include "version.h"
 #include "lpc10codec.h"
+#include "speexcodec.h"
+
 #include "mscodecs.h"
+#include "opalvxml.h"
 #include "main.h"
 
 PCREATE_PROCESS(OpenAm);
@@ -257,11 +328,7 @@ PCREATE_PROCESS(OpenAm);
 #define	DEFAULT_MSG_LIMIT	30
 #define DEFAULT_CALL_LOG	"call_log.txt"
 
-#define	G7231_BLOCK_SIZE	24
 #define	G7231_SAMPLES_PER_BLOCK	240
-#define	G7231_BANDWIDTH		63
-#define	G7231_SILENCE_FLAG	0x02
-#define	G7231_SILENCE_LENGTH	0x04
 
 #define CHECK_PCM   1
 #define CHECK_G7231 2
@@ -274,7 +341,7 @@ static PFilePath logFilename = DEFAULT_CALL_LOG;
 
 PString G7231Ext	= ".g723";
 PString WAVExt          = ".wav";
-static PString PCMExt   = ".sw";
+PString PCMExt          = ".sw";
 
 static void LogMessage(const PString & str)
 {
@@ -342,11 +409,13 @@ void OpenAm::Main()
              "-gsm."                 "-no-gsm."
              "-gsmmessage:"          "-no-gsmmessage."
              "h-help."
+             "H-hangup."             "-no-hangup."
              "i-interface:"          "-no-interface."
              "k-kill."               "-no-kill."
              "l-limit:"              "-no-limit."
              "-listenport:"          "-no-listenport."
              "-lpc10message:"        "-no-lpc10message."
+             "-speexmessage:"        "-no-speexmessage."
              "m-message:"            "-no-message."
              "-no-recordg7231."
 #if PTRACING
@@ -355,6 +424,7 @@ void OpenAm::Main()
              "P-prefer:"
              "-pcm."                 "-no-pcm."
              "-pcmmessage:"          "-no-pcmmessage."
+             "-port:"
              "q-quicknet:"           "-no-quicknet:"
              "r-run:"                "-no-run."
 	     "-recordraw."
@@ -389,9 +459,12 @@ void OpenAm::Main()
             "  --g711message fn    : Set outgoing message for G711 codec to fn\n"
             "  --gsmmessage fn     : Set outgoing message for GSM codec to fn\n"
             "  --lpc10message fn   : Set outgoing message for LPC10 codec to fn\n"
+            "  --speexmessage fn   : Set outgoing message for Speex codec to fn\n"
+
             "  --recordraw         : Record PCM audo in raw files (.sw) instead of .wav\n"
             "  -r --run cmd        : Run this command after each recorded message\n"
             "  -k --kill           : Kill recorded files after user command\n"
+            "  -H --hangup         : hangup after playing message\n"
             "  -u --username str   : Set the local endpoint name to str\n"
             "  -i --interface ip   : Bind to a specific interface\n"
             "  --listenport port   : Listen on a specific port\n"
@@ -456,6 +529,9 @@ void OpenAm::Main()
       cout << "WARNING: recorded files will be deleted even though no run command is present" << endl;
     flags |= MyH323EndPoint::DeleteAfterRecord;
   }
+
+  if (args.HasOption('H'))
+    flags |= MyH323EndPoint::HangupAfterPlay;
   
   MyH323EndPoint endpoint(callLimit, runCmd, dir, flags);
 
@@ -470,7 +546,7 @@ void OpenAm::Main()
   // start the H.323 listener
   H323ListenerTCP * listener;
   PIPSocket::Address interfaceAddress(INADDR_ANY);
-  WORD listenPort = H323ListenerTCP::DefaultSignalPort;
+  WORD listenPort = H323EndPoint::DefaultTcpPort;
 
   if (args.HasOption("listenport"))
     listenPort = (WORD)args.GetOptionString("listenport").AsInteger();
@@ -514,135 +590,6 @@ void OpenAm::Main()
 }
 
 
-///////////////////////////////////////////////////////////////
-//  : H323AudioCapability(12*G7231_BLOCK_SIZE, 4*G7231_BLOCK_SIZE)
-
-G7231_File_Capability::G7231_File_Capability()
-  : H323AudioCapability(8, 4)
-{
-}
-
-BOOL G7231_File_Capability::OnSendingPDU(H245_AudioCapability & cap, unsigned packetSize) const
-{
-  // set the choice to the correct type
-  cap.SetTag(GetSubType());
-
-  // get choice data
-  H245_AudioCapability_g7231 & g7231 = cap;
-
-  // max number of audio frames per PDU we want to send
-  g7231.m_maxAl_sduAudioFrames = packetSize; 
-
-  // no silence suppression
-  g7231.m_silenceSuppression = FALSE;
-
-  return TRUE;
-}
-
-BOOL G7231_File_Capability::OnReceivedPDU(const H245_AudioCapability & cap, unsigned & packetSize)
-{
-  const H245_AudioCapability_g7231 & g7231 = cap;
-  packetSize = g7231.m_maxAl_sduAudioFrames;
-  return TRUE;
-}
-
-PObject * G7231_File_Capability::Clone() const
-{
-  return new G7231_File_Capability();
-}
-
-H323Codec * G7231_File_Capability::CreateCodec(H323Codec::Direction direction) const
-{
-  return new G7231_File_Codec(direction);
-}
-
-///////////////////////////////////////////////////////////////
-
-G7231_File_Codec::G7231_File_Codec(Direction dir)
-  : H323AudioCodec("G.723.1", dir)
-{
-}
-
-
-int G7231_File_Codec::GetFrameLen(int val)
-{
-  static const int frameLen[] = { 24, 20, 4, 1 };
-  return frameLen[val & 3];
-}
-
-BOOL G7231_File_Codec::Read(BYTE * buffer, unsigned & length, RTP_DataFrame &)
-{
-  if (rawDataChannel == NULL)
-    return FALSE;
-    
-  if (!rawDataChannel->Read(buffer, 1)) {
-    cerr << "Read failed" << endl;
-    return FALSE;
-  }
-  if (rawDataChannel->GetLastReadCount() == 0) {
-    cerr << "Read count is zero" << endl;
-    return FALSE;
-  }
-
-  int readLen = G7231_File_Codec::GetFrameLen(buffer[0]);
-
-  if (readLen > 0) {
-    if (!rawDataChannel->Read(buffer+1, readLen-1)) {
-      cerr << "Read failed" << endl;
-      return FALSE;
-    }
-    if (rawDataChannel->GetLastReadCount() == 0) {
-      cerr << "Read count is zero" << endl;
-      return FALSE;
-    }
-  }
-
-  length = readLen;
-  return TRUE;
-}
-
-BOOL G7231_File_Codec::Write(const BYTE * buffer, unsigned length, const RTP_DataFrame & /* rtp */, unsigned & writtenLength)
-{
-  static const BYTE silence[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-                                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-                                  0, 0, 0, 0};
-  // If the length is zero, output silence to the file..
-  if (length == 0) {
-    PTRACE(1,"G.723.1 zero length frame");
-    writtenLength = 0;
-    return rawDataChannel->Write(silence, 24);
-  }
-
-  int writeLen;
-  switch (buffer[0]&3) {
-    case 0:
-      writeLen = 24;
-      break;
-    case 1:
-      writeLen = 20;
-      break;
-    case 2:
-      // Windows Media Player cannot play 4 byte SID (silence) frames.
-      // So write out a 24 byte frame of silence instead.
-      PTRACE(1,"Replacing SID with 24 byte frame");
-      writtenLength = 4;
-      return rawDataChannel->Write(silence, 24);
-//    writeLen = 4;
-//    break;
-    default :
-      writeLen = 1;
-      break;
-  }
-
-  PTRACE(1,"G.723.1 frame length = " <<writeLen);
-
-  writtenLength = writeLen;
-
-  return rawDataChannel->Write(buffer, writeLen);
-}
-
-unsigned G7231_File_Codec::GetBandwidth() const      { return G7231_BANDWIDTH; }
-      
 ///////////////////////////////////////////////////////////////
 
 
@@ -694,7 +641,8 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
   }
 
   if (!g7231Ogm.IsEmpty()) {
-    if (!PFile::Exists(g7231Ogm)) {
+    // check if the file exists. (do not check if filename contains %s)
+    if ((g7231Ogm.Find("%s") == P_MAX_INDEX) && !PFile::Exists(g7231Ogm)) {
       cout << "warning: cannot open G723.1 OGM file \"" << g7231Ogm << "\"" << endl;
       g7231Ogm = "";
     } 
@@ -714,15 +662,15 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
     pcmOgm = args.GetOptionString("pcmmessage");
   }
   else if (args.HasOption('m')) {
-    if (PFile::Exists(args.GetOptionString('m'))) {
-      pcmOgm = args.GetOptionString('m');
-    }
-    else if (PFile::Exists(args.GetOptionString('m') + WAVExt)) {
-      pcmOgm = args.GetOptionString('m') + WAVExt;
-    }
-    else if (PFile::Exists(args.GetOptionString('m') + PCMExt)) {
-      pcmOgm = args.GetOptionString('m') + PCMExt;
-    }
+      if (PFile::Exists(args.GetOptionString('m'))) {
+        pcmOgm = args.GetOptionString('m');
+      }
+      else if (PFile::Exists(args.GetOptionString('m') + WAVExt)) {
+        pcmOgm = args.GetOptionString('m') + WAVExt;
+      }
+      else if (PFile::Exists(args.GetOptionString('m') + PCMExt)) {
+        pcmOgm = args.GetOptionString('m') + PCMExt;
+      }
   }
 
 
@@ -731,6 +679,7 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
   gsmOgm   = pcmOgm;
   g711Ogm  = pcmOgm;
   lpc10Ogm = pcmOgm;
+  speexOgm = pcmOgm;
 
 
   // We can set the filename for specific codecs.
@@ -743,9 +692,12 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
   if (args.HasOption("lpc10message")) 
     lpc10Ogm = args.GetOptionString("lpc10message");
 
+  if (args.HasOption("speexmessage")) 
+    speexOgm = args.GetOptionString("speexmessage");
+
   // Check GSM OGM message
   if (!gsmOgm.IsEmpty()) {
-    if (!PFile::Exists(gsmOgm)) {
+    if ((gsmOgm.Find("%s") == P_MAX_INDEX) && !PFile::Exists(gsmOgm)) {
       cout << "warning: cannot open GSM OGM file \"" << gsmOgm << "\"" << endl;
       gsmOgm = "";
     }
@@ -758,7 +710,7 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
 
   // Check G.711 OGM message 
   if (!g711Ogm.IsEmpty()) {
-    if (!PFile::Exists(g711Ogm)) {
+    if ((g711Ogm.Find("%s") == P_MAX_INDEX) && !PFile::Exists(g711Ogm)) {
       cout << "warning: cannot open G711 OGM file \"" << g711Ogm << "\"" << endl;
       g711Ogm = "";
     }
@@ -771,7 +723,7 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
 
   // Check LPC10 OGM message 
   if (!lpc10Ogm.IsEmpty()) {
-    if (!PFile::Exists(lpc10Ogm)) {
+    if ((lpc10Ogm.Find("%s") == P_MAX_INDEX) && !PFile::Exists(lpc10Ogm)) {
       cout << "warning: cannot open LPC10 OGM file \"" << lpc10Ogm << "\"" << endl;
       lpc10Ogm = "";
     }
@@ -782,9 +734,25 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
     cout << "Using \"" << lpc10Ogm << "\" as LPC10 outgoing message\n";
   }
 
+  // Check Speex OGM message 
+  if (!speexOgm.IsEmpty()) {
+    // check if the file exists. (do not check if filename contains %s)
+    if ((speexOgm.Find("%s") == P_MAX_INDEX) && !PFile::Exists(speexOgm)) {
+      cout << "warning: cannot open Speex OGM file \"" << speexOgm << "\"" << endl;
+      speexOgm = "";
+    }
+  } 
+  if (speexOgm.IsEmpty()) 
+    cout << "No Speex outgoing message set\n";
+  else {
+    cout << "Using \"" << speexOgm << "\" as Speex outgoing message\n";
+  }
+
 
   if (g7231Ogm.IsEmpty() && gsmOgm.IsEmpty() && g711Ogm.IsEmpty()
-                         && lpc10Ogm.IsEmpty()) {
+                         && lpc10Ogm.IsEmpty() 
+			 && speexOgm.IsEmpty()
+			 ) {
     cerr << "Must specify at least one outgoing message" << endl;
     return FALSE;
   }
@@ -807,6 +775,13 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
   if (!lpc10Ogm.IsEmpty())
     SetCapability(0, 0, new H323_LPC10Capability(*this));
 
+  if (!speexOgm.IsEmpty())
+    SetCapability(0, 0, new SpeexNarrow2AudioCapability());
+    SetCapability(0, 0, new SpeexNarrow3AudioCapability());
+    SetCapability(0, 0, new SpeexNarrow4AudioCapability());
+    SetCapability(0, 0, new SpeexNarrow5AudioCapability());
+    SetCapability(0, 0, new SpeexNarrow6AudioCapability());
+
   capabilities.Remove(args.GetOptionString('D').Lines());
   capabilities.Reorder(args.GetOptionString('P').Lines());
 
@@ -1094,7 +1069,9 @@ H323Connection::AnswerCallResponse
   return AnswerCallNow;
 }
 
-BOOL MyH323Connection::OpenAudioChannel(BOOL isEncoding, unsigned /* bufferSize */, H323AudioCodec & codec)
+BOOL MyH323Connection::OpenAudioChannel(BOOL isEncoding, 
+                                        unsigned /* bufferSize */, 
+                                        H323AudioCodec & codec)
 {
   codec.SetSilenceDetectionMode(H323AudioCodec::NoSilenceDetection);
   PStringStream codecName;
@@ -1121,6 +1098,9 @@ BOOL MyH323Connection::OpenAudioChannel(BOOL isEncoding, unsigned /* bufferSize
   } else if (codec.IsDescendant(H323_LPC10Codec::Class())) {
     ogm   = ep.GetLPC10OGM();
     isPCM = TRUE;
+  } else if (codec.IsDescendant(SpeexCodec::Class())) {
+    ogm   = ep.GetSPEEXOGM();
+    isPCM = TRUE;
   } else {
     cerr << "Unknown codec \"" << codecName << endl;
     return FALSE;
@@ -1132,15 +1112,15 @@ BOOL MyH323Connection::OpenAudioChannel(BOOL isEncoding, unsigned /* bufferSize
   if ((recordFile == NULL) && (isEncoding == FALSE)) {
     if (isPCM) {
       if (ep.GetRecordWav() == TRUE)
-	recordFn = ep.GetDirectory() + (basename + ".wav");
+	      recordFn = ep.GetDirectory() + (basename + ".wav");
       else
-	recordFn = ep.GetDirectory() + (basename + ".sw");
+	      recordFn = ep.GetDirectory() + (basename + ".sw");
       recordFile = new PCM_RecordFile  (*this, recordFn, ep.GetCallLimit());
     } else {
       if (ep.GetRecordWav() == TRUE)
-	recordFn = ep.GetDirectory() + (basename + ".wav");
+	      recordFn = ep.GetDirectory() + (basename + ".wav");
       else
-	recordFn = ep.GetDirectory() + (basename + ".g723");
+	      recordFn = ep.GetDirectory() + (basename + ".g723");
       recordFile = new G7231_RecordFile(*this, recordFn, ep.GetCallLimit());
     }
   }
@@ -1153,6 +1133,13 @@ BOOL MyH323Connection::OpenAudioChannel(BOOL isEncoding, unsigned /* bufferSize
   }
 
   if (isEncoding) {
+
+    if (ep.GetHangupAfterPlay())
+      ogmChannel->SetPlayOnce();
+
+    if (ogm.Find("%s"))
+      ogm.Replace("%s", e164Number);
+
     transmitCodecName = codecName;
     if (!StartMenu(0)) {
       if (!PFile::Exists(ogm))
@@ -1351,13 +1338,9 @@ PCM_OGMChannel::PCM_OGMChannel(MyH323Connection & _conn)
   recordTrigger = FALSE;
   hangupTrigger = FALSE;
   closed        = FALSE;
-
-  headRoom = 0;
-  firstRead = TRUE;
+  playOnce      = FALSE;
 
   frameLen = frameOffs = 0;
-
-  frameBuffer = NULL;       // size the buffer when we know what the codec wants
 }
 
 
@@ -1384,30 +1367,6 @@ BOOL PCM_OGMChannel::IsWAVFileValid(PWAVFile *chan) {
 }
 
 
-void PCM_OGMChannel::FrameDelay(int offs)
-{
-  if (firstRead)
-    firstRead = FALSE;
-  else {
-
-    // subtract actual time taken for previous frame
-    PTime now;
-    PTimeInterval interval = now - lastReadTime;
-    headRoom -= (int)interval.GetMilliSeconds();
-
-    // save time of read so we can calculate delay next time
-    lastReadTime = PTime();
-
-    // if we have acculated too much headroom, then reduce it
-    if (headRoom >= Max_Headroom) {
-      int desiredDelay = headRoom - Min_Headroom;
-      PThread::Current()->Sleep(desiredDelay);
-    }
-  }
-
-  headRoom += offs;
-}
-
 BOOL PCM_OGMChannel::Read(void * buffer, PINDEX amount)
 {
   PWaitAndSignal mutex(chanMutex);
@@ -1453,7 +1412,7 @@ BOOL PCM_OGMChannel::Read(void * buffer, PINDEX amount)
             PTRACE(1, "Cannot open file \"" << chan->GetName() << "\"");
             delete chan;
           } else {
-	    if (!IsWAVFileValid(chan) ){
+	          if (!IsWAVFileValid(chan) ){
               PTRACE(1, chan->GetName() << " is not a valid wav file");
               delete chan;
               cerr << "wave file is invalid" << endl;
@@ -1499,12 +1458,14 @@ BOOL PCM_OGMChannel::Read(void * buffer, PINDEX amount)
         silentCount = 5;   // always do 5 frames of silence after every file
   
         // hangup if required
-        if (hangupTrigger) 
+        if (hangupTrigger || playOnce) 
           conn.ClearCall();
   
         // trigger record if required
-        else if (recordTrigger) 
-          conn.StartRecording();
+        else if (recordTrigger) {
+          if ((playQueue.GetSize() == 0) && (GetBaseReadChannel() == NULL))
+            conn.StartRecording();
+	      }
    
         // no silence
         doSilence = TRUE;
@@ -1591,14 +1552,23 @@ BOOL PCM_OGMChannel::AdjustFrame(void * buffer, PINDEX amount)
 
 void PCM_OGMChannel::Synchronise(PINDEX amount)
 {
-  FrameDelay(amount / 16);
+  ogm_delay.Delay(amount / 16);
 }
 
 BOOL PCM_OGMChannel::ReadFrame(PINDEX amount)
 {
   frameOffs = 0;
   frameLen  = amount;
-  return PIndirectChannel::Read(frameBuffer.GetPointer(), frameLen);
+
+  BOOL result = PIndirectChannel::Read(frameBuffer.GetPointer(), frameLen);
+
+  // if we did not read a full frame of audio, fill the end of the
+  // frame with zeros.
+  PINDEX count = GetLastReadCount();
+  if (count < frameLen)
+    memset(frameBuffer.GetPointer()+count, 0, frameLen-count);
+
+  return result;
 }
 
 void PCM_OGMChannel::CreateSilenceFrame(PINDEX amount)
@@ -1617,7 +1587,7 @@ G7231_OGMChannel::G7231_OGMChannel(MyH323Connection & conn)
 
 void G7231_OGMChannel::Synchronise(PINDEX /*amount*/)
 {
-  FrameDelay(30);
+  ogm_delay.Delay(30);
 }
 
 BOOL G7231_OGMChannel::ReadFrame(PINDEX /*amount*/)
@@ -1646,47 +1616,3 @@ BOOL G7231_OGMChannel::IsWAVFileValid(PWAVFile *chan) {
 }
 
 ///////////////////////////////////////////////////////////////
-
-#define MIN_HEADROOM    30
-#define MAX_HEADROOM    60
-
-AudioDelay::AudioDelay()
-{
-  firstTime = TRUE;
-  error = 0;
-}
-
-void AudioDelay::Restart()
-{
-  firstTime = TRUE;
-}
-
-BOOL AudioDelay::Delay(int frameTime)
-{
-  if (firstTime) {
-    firstTime = FALSE;
-    previousTime = PTime();
-    return TRUE;
-  }
-
-  error += frameTime;
-
-  PTime now;
-  PTimeInterval delay = now - previousTime;
-  error -= (int)delay.GetMilliSeconds();
-  previousTime = now;
-
-  if (error > 0)
-#ifdef P_LINUX
-    usleep(error * 1000);
-#else
-    PThread::Current()->Sleep(error);
-#endif
-
-  return error <= -frameTime;
-
-  //if (headRoom > MAX_HEADROOM)
-  //  PThread::Current()->Sleep(headRoom - MIN_HEADROOM);
-}
-
-// End of File ///////////////////////////////////////////////////////////////
